from gai.common.utils import get_lib_config
from gai.common.http_utils import http_post
from gai.lib.ClientBase import ClientBase
from gai.common.logging import getLogger
from pydantic import BaseModel
logger = getLogger(__name__)

class Transcription(BaseModel):
    text:str

class STTClient(ClientBase):

    def __init__(self, type, config_path=None):
        super().__init__(category_name="stt",type=type,config_path=config_path)

    def __call__(self, generator_name=None, file=None, file_path=None):

        if generator_name:
            raise Exception("Customed generator_name not supported.")

        if self.type=="openai":
            return self.openai_whisper(file=file)

        if self.type=="gai":

            if file_path:
                with open(file_path, "rb") as f:
                    data = f.read()
                files = {
                    "file": (file_path, data)
                }

                url = self._get_gai_url()
                response = http_post(url, files=files)
                return Transcription(text=response.json()["text"])

            if file:
                files = {
                    "file": (file.name, file.read())
                }
                url = self._get_gai_url()
                response = http_post(url, files=files)
                return Transcription(text=response.json()["text"])

            raise Exception("No file provided")

    def openai_whisper(self, **model_params):
        import os
        import openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()

        if "file" not in model_params:
            raise Exception("Missing file parameter")

        file = model_params["file"]

        # If file is a bytes object, we need to write it to a temporary file then pass the file object to the API
        if isinstance(file, bytes):
            import tempfile
            with tempfile.NamedTemporaryFile(suffix='.wav') as temp:
                temp.write(file)
                temp.flush()
                temp.seek(0)
                model_params["file"] = temp.file
                response = client.audio.transcriptions.create(
                    model='whisper-1', **model_params)
        else:
            response = client.audio.transcriptions.create(
                model='whisper-1', **model_params)

        return response



from gai.lib.ttt.ChunkWrapper import ChunkWrapper
from gai.common.utils import get_lib_config
from gai.common.http_utils import http_post
from gai.common.image_utils import base64_to_imageurl
from gai.lib.ttt.OpenAIChunkWrapper import OpenAIChunkWrapper
from gai.common.generators_utils import chat_string_to_list
from gai.lib.ClientBase import ClientBase
from gai.common.logging import getLogger
logger = getLogger(__name__)

class ITTClient(ClientBase):

    def __init__(self, type, config_path=None):
        super().__init__(category_name="itt",type=type,config_path=config_path)

    def __call__(self, generator_name=None, messages=None, stream=True, **generator_params):
        if generator_name:
            raise Exception("Customed generator_name not supported.")

        if self.type == "openai":
            return self.openai_vision(messages=messages, stream=stream, **generator_params)
        if self.type == "gai":
            return self.api(messages=messages, stream=stream, **generator_params)

        raise Exception("Generator type not supported.")
        

    def api(self, messages=None, stream=True, **generator_params):
        if not messages:
            raise Exception("Messages not provided")

        data = {
            "messages": messages,
            "stream": stream,
            **generator_params
        }

        def streamer(response):
            for chunk in response.iter_lines():
                yield ChunkWrapper(chunk)

        response = http_post(self._get_gai_url(), data)

        if not stream:
            response.decode = lambda: response.json(
            )["choices"][0]["message"]["content"]
            return response
        return streamer(response)

    def openai_vision(self, messages=None, stream=True, **generator_params):
        import os
        import openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield OpenAIChunkWrapper(chunk)

        model = "gpt-4-vision-preview"
        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        response = client.chat.completions.create(
            model=model,
            messages=messages,
            stream=stream,
            **generator_params
        )

        if not stream:
            response.decode = lambda: response.choices[0].message.content
            return response
        return streamer(response)



import os, sys, re, time
import json
from os.path import dirname
import shutil
from gai.common import constants
import yaml

def init():
    with open(os.path.expanduser(constants.GAIRC), "w") as file:
        file.write(json.dumps({
            "app_dir": "~/gai"
        }, indent=4))
    config_dir=dirname(dirname(dirname(__file__)))
    config_path=os.path.join(config_dir, 'gai.json')
    os.makedirs(os.path.expanduser("~/gai/models"), exist_ok=True)
    shutil.copy(config_path, os.path.expanduser("~/gai"))

# Get JSON FROM ~/.gairc
def get_rc():
    if (not os.path.exists(os.path.expanduser(constants.GAIRC))):
        init()
    with open(os.path.expanduser(constants.GAIRC), 'r') as f:
        return json.load(f)

# Get "app_dir" from ~/.gairc
def get_app_path():
    rc = get_rc()
    app_dir=os.path.abspath(os.path.expanduser(rc["app_dir"]))
    return app_dir

# Get ~/.gai/gai.json
def get_gen_config(file_path=None):
    app_dir=get_app_path()
    gen_config_path = os.path.join(app_dir, 'gai.json')
    if file_path:
        gen_config_path = file_path
    with open(gen_config_path, 'r') as f:
            return json.load(f)

# Get ~/.gai/gai.yml
def get_lib_config(file_path=None):
    app_dir=get_app_path()
    lib_config_path = os.path.join(app_dir, 'gai.yml')
    if file_path:
        lib_config_path = file_path
    with open(lib_config_path, 'r') as f:
        return yaml.load(f, Loader=yaml.FullLoader)

def this_dir(file):
    return os.path.dirname(os.path.realpath(file))

# Create ~/.gaiaio/cache
def mkdir_cache():
    cache_dir = os.path.expanduser('~/.gaiaio/cache')
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir

def is_url(s):
    return re.match(r'^https?:\/\/.*[\r\n]*', s) is not None

def sha256(text):
    import hashlib
    return hashlib.sha256(text.encode('utf-8')).hexdigest()


def timestamp():
    return int(time.time() * 1000)

def find_url_in_text(text):
    url_pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    urls = re.findall(url_pattern, text)
    return urls

def clean_string(s):
    if s is None:
        return ''
    return re.sub(r'\s+', ' ', s)

def find_site_packages_path(virtual_env_name):
    # Extracting the Python version
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"

    # Constructing the path pattern to the virtual environment's site-packages directory
    site_packages_path = os.path.expanduser(f"~/miniconda/envs/{virtual_env_name}/lib/python{python_version}/site-packages")
    return site_packages_path

def find_egg_link(virtual_env_name, package_name):
    site_packages_path = find_site_packages_path(virtual_env_name)
    egg_link_file = os.path.join(site_packages_path, f"{package_name}.egg-link")
    if os.path.exists(egg_link_file):
        return egg_link_file
    else:
        return None

def find_project_path(virtual_env_name, package_name):
    egg_link_file = find_egg_link(virtual_env_name, package_name)
    if egg_link_file is None:
        return None
    else:
        with open(egg_link_file) as f:
            project_path = f.readline().strip()
            return project_path



from gai.lib.ttt.ChunkWrapper import ChunkWrapper
from gai.lib.ttt.OpenAIChunkWrapper import OpenAIChunkWrapper
from gai.lib.ttt.AnthropicChunkWrapper import AnthropicChunkWrapper
from gai.common.utils import get_lib_config
from gai.common.http_utils import http_post
from gai.common.generators_utils import chat_string_to_list, chat_list_to_string
from gai.common.errors import ApiException
from gai.common.logging import getLogger
logger = getLogger(__name__)
import json,os
from gai.lib.ClientBase import ClientBase
from dotenv import load_dotenv
load_dotenv()

class TTTClient(ClientBase):

    def __init__(self, type, config_path=None):
        super().__init__(category_name="ttt", type=type, config_path=config_path)

    def __call__(self, 
                 messages:str|list, 
                 stream:bool=True, 
                 max_new_tokens:int=None, 
                 max_tokens:int=None, 
                 temperature:float=None, 
                 top_p:float=None, 
                 top_k:float=None,
                 schema:dict=None,
                 tools:list=None,
                 tool_choice:str=None):

        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        # If generator is openai-gpt4 or claude2-100k, 
        # use the respective API client instead of generator service.
        if self.type == "openai":
            return self.gpt_4(messages, 
                              stream=stream, 
                              max_tokens=max_tokens, 
                              temperature=temperature, 
                              top_p=top_p, 
                              top_k=top_k, 
                              schema=schema,
                              tool_choice=tool_choice,
                              tools=tools)
        if self.type == "gai":
            return self.api(messages, 
                            stream=stream, 
                            max_new_tokens=max_new_tokens, 
                            temperature=temperature, 
                            top_p=top_p, 
                            top_k=top_k, 
                            schema=schema,
                            tool_choice=tool_choice,
                            tools=tools)

        raise Exception("Generator type not supported.")


    def api(self, messages:list, stream:bool, max_new_tokens:int, temperature:float, top_p:float, top_k:float, schema:dict, tools:list, tool_choice:str):
        #logger.debug(f'TTTClient.api: messages={messages}')

        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        if not messages:
            raise Exception("Messages not provided")


        data = { 
            "messages": messages,
            "stream": stream,
            "max_new_tokens": max_new_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "top_k": top_k,
            "schema": schema,
            "tools": tools,
            "tool_choice": tool_choice
        }

        def streamer(response):
            for chunk in response.iter_lines():
                yield ChunkWrapper(chunk)

        try:
            url = self._get_gai_url()
            response = http_post(url, data)
        except ApiException as he:

            # Switch to long context
            if he.code == "context_length_exceeded":
                try:
                    url = self._get_gai_url("url")
                    response = http_post(url, data)
                except Exception as e:
                    logger.error(f"TTTClient.api: gaigen error={e}")
                    raise e
            else:
                raise he
        except Exception as e:
            logger.error(f"TTTClient.api: gaigen error={e}")
            raise e

        if not stream:
            if response.json()["choices"][0]["message"]["tool_calls"]:
                response.decode = lambda: {
                    "type":"function",
                    "name": response.json()["choices"][0]["message"]["tool_calls"][0]["function"]["name"],
                    "arguments": response.json()["choices"][0]["message"]["tool_calls"][0]["function"]["arguments"],
                }
            elif response.json()["choices"][0]["message"]["content"]:
                response.decode = lambda: {
                    "type": "content",
                    "content": response.json()["choices"][0]["message"]["content"]
                }
            return response
        return streamer(response)


    # Call GPT-4 API
    def gpt_4(self, messages:list, stream:bool, max_tokens:int, temperature:float, top_p:float, top_k:float, schema:dict, tools:list,tool_choice:str):
        import os

        from openai import OpenAI

        # Get API KEY
        OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", None)
        if not OPENAI_API_KEY:
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        client = OpenAI(api_key=OPENAI_API_KEY)

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield OpenAIChunkWrapper(chunk)

        model = "gpt-4o"

        response_format = None
        if schema:
            response_format={
                "type":"json_object"
            }
            messages[-2]["content"] += f"Format your response in this json schema: {schema}"
        
        if not tools:
            tool_choice=None
        else:
            if isinstance(tools,dict):
                tools = [tool for tool in tools.values() if tool]

        response = client.chat.completions.create(
            model=model,
            messages=messages,
            stream=stream,
            max_tokens=max_tokens,
            response_format=response_format,
            temperature=temperature,
            top_p=top_p,
            tool_choice=tool_choice,
            tools=tools,
        )

        if not stream:
            if response.choices[0].message.tool_calls:
                response.decode = lambda: {
                    "type":"function",
                    "name": response.choices[0].message.tool_calls[0].function.name,
                    "arguments": response.choices[0].message.tool_calls[0].function.arguments,
                }
            elif response.choices[0].message.content:
                response.decode = lambda: {
                    "type": "content",
                    "content": response.choices[0].message.content
                }
            return response


        if not stream:
            response.decode = lambda: response.choices[
                0].message.content if response.choices[0].message.content else ""
            return response
        return streamer(response)

    # Call Claude API
    def claude_2(self, messages:list, stream:bool, **generator_params):
        # import os
        # from anthropic import Anthropic
        # from dotenv import load_dotenv
        # load_dotenv()
        # if not os.environ.get("ANTHROPIC_API_KEY"):
        #     raise Exception(
        #         "ANTHROPIC_API_KEY not found in environment variables")
        # client = Anthropic()

        from anthropic import Anthropic
        ANTHROPIC_API_KEY = generator_params.pop("ANTHROPIC_API_KEY", None)
        if not ANTHROPIC_API_KEY:
            raise Exception("ANTHROPIC_API_KEY not provided in generator_params")
        client = Anthropic(api_key=ANTHROPIC_API_KEY)

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield AnthropicChunkWrapper(chunk)

        model = "claude-2"
        message = messages
        if isinstance(messages, list):
            message = chat_list_to_string(messages)
        prompt_template = "\n\nHuman: {message}\n\nAssistant:"
        messages = prompt_template.format(message=message)

        # in case max_tokens_to_sample is not provided, set it to 200
        max_tokens_to_sample = generator_params.pop("max_tokens_to_sample", 200)
        generator_params["max_tokens_to_sample"] = max_tokens_to_sample
        response = client.completions.create(
            model=model,
            prompt=messages,
            stream=stream,
            **generator_params
        )

        if not stream:
            response.decode = lambda: response.completion
            return response
        return streamer(response)



from enum import Enum
from PIL import Image
from io import BytesIO
from gai.common.http_utils import http_post, http_get
from gai.lib.ClientBase import ClientBase
from gai.common.logging import getLogger
logger = getLogger(__name__)
import json, base64
from gai.common.image_utils import resize_image

class TTIOutputType(Enum):
    BYTES="bytes"
    DATA_URL="data_url"
    IMAGE="image"

class TTIClient(ClientBase):

    def __init__(self, type,config_path=None):
        super().__init__(category_name="tti",type=type,config_path=config_path)

    def __call__(self, 
                 prompt:str, 
                 negative_prompt:str=None,
                 width:int=512,
                 height:int=512,
                 steps:int=0,
                 output_type:TTIOutputType = TTIOutputType.BYTES
                 ):
        if not prompt:
            raise Exception("The parameter 'input' is required.")

        if self.type == "openai":
            if steps == 0:
                steps = 1
            return self.openai_tti(prompt=prompt,
                                width=width,
                                height=height,
                                n=steps,
                                output_type=output_type)
        
        if self.type == "gai":
            if steps == 0:
                steps = 10
            return self.gai_tti(prompt=prompt,
                                negative_prompt=negative_prompt, 
                                width=width,
                                height=height,
                                steps=steps,
                                output_type=output_type)
        raise Exception("Generator type not supported.")

    def gai_tti(self, 
                prompt:str,
                negative_prompt:str,
                width:int,
                height:int,
                steps:int,
                output_type:TTIOutputType = TTIOutputType.BYTES
                ):
        negative_prompt = negative_prompt or "ugly, duplicate, morbid, mutilated, out of frame, extra fingers, mutated hands, poorly drawn hands, poorly drawn face, mutation, deformed, ugly, blurry, bad anatomy, bad proportions, extra limbs, cloned face, out of frame, ugly, extra limbs, bad anatomy, gross proportions, malformed limbs, missing arms, missing legs, extra arms, extra legs, mutated hands, fused fingers, too many fingers, long neck, extra head, cloned head, extra body, cloned body, watermark. extra hands, clone hands, weird hand, weird finger, weird arm, (mutation:1.3), (deformed:1.3), (blurry), (bad anatomy:1.1), (bad proportions:1.2), out of frame, ugly, (long neck:1.2), (worst quality:1.4), (low quality:1.4), (monochrome:1.1), text, signature, watermark, bad anatomy, disfigured, jpeg artifacts, 3d max, grotesque, desaturated, blur, haze, polysyndactyly"
        data = {
            "prompt": prompt,
            "negative_prompt": negative_prompt,
            "width": width,
            "height": height,
            "steps": steps
        }
        response = http_post(self._get_gai_url(), data)
        base64_img = json.loads(response.content.decode("utf-8"))["images"][0]
        image_data = base64.b64decode(base64_img)
        
        if TTIOutputType(output_type) == TTIOutputType.DATA_URL:
            base64_encoded_data = base64.b64encode(image_data).decode('utf-8')
            data_url = f"data:image/jpeg;base64,{base64_encoded_data}"
            return data_url
        elif TTIOutputType(output_type) == TTIOutputType.IMAGE:
            return Image.open(BytesIO(image_data))
        elif TTIOutputType(output_type) == TTIOutputType.BYTES:
            return image_data
        else:
            raise Exception("Output type not supported.")

    # OpenAI default image size is 1024x1024 but we will convert to 512x512
    def openai_tti(self, 
                prompt:str,
                width:int,
                height:int,
                n: int,
                output_type:TTIOutputType = TTIOutputType.BYTES
                ):
        import os, openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()
        response = client.images.generate(
            model='dall-e-3',
            prompt=prompt,
            size=f"1024x1024",
            quality="standard",
            n=n
            )
        response = http_get(response.data[0].url)
        image_data = response.content
        image_data=resize_image(image_data, width, height)

        if TTIOutputType(output_type) == TTIOutputType.DATA_URL:
            binary_data = image_data
            base64_encoded_data = base64.b64encode(binary_data).decode('utf-8')
            data_url = f"data:image/jpeg;base64,{base64_encoded_data}"
            return data_url
        elif TTIOutputType(output_type) == TTIOutputType.IMAGE:
            return Image.open(BytesIO(image_data))
        elif TTIOutputType(output_type) == TTIOutputType.BYTES:
            return image_data
        else:
            raise Exception("Output type not supported.")




from gai.lib.ttt.TTTClient import TTTClient
from gai.lib.STTClient import STTClient
from gai.lib.TTSClient import TTSClient
from gai.lib.ITTClient import ITTClient
from gai.lib.TTIClient import TTIClient
from gai.common.utils import get_lib_config
from PIL import Image
import base64
from io import BytesIO

class GGG:

    client = None

    def __init__(self,category, type, config_path=None):
        self.config_path = config_path
        if config_path:
            self.config = get_lib_config(config_path)
        else:
            self.config = get_lib_config()
        self.category = category.lower()
        self.type = type.lower()

    def __call__(self, **model_params):
        key = f"{self.category}-{self.type}" ### eg. ttt-gai or ttt-openai

        # Check if category/type is in the config
        category_config = self.config["generators"].get(key,None)

        if not category_config:
            raise Exception(f"Unknown category: {self.category}")
        
        # TTT: Text-to-Text Generation up to 8096 tokens
        if self.category == "ttt":
            self.client = TTTClient(type=self.type,config_path=self.config_path)
            return self.client(**model_params)

        # TTT Long: > 8096 tokens. Uses the same client as TTT, but with a different generator
        if self.category == "ttt-long":
            self.client = TTTClient(type=self.type,config_path=self.config_path)
            return self.client(**model_params)

        # STT: There are 3 ways to provide audio data: file_path, file, and audio        
        if self.category == "stt":
            self.client = STTClient(type=self.type,config_path=self.config_path)

            if "file_path" in model_params:
                file_path = model_params.pop("file_path", None)
                return self.client(file=open(file_path, "rb"), **model_params)

            if "file" in model_params:
                file = model_params.pop("file", None)
                return self.client(file=file, **model_params)

            return self.client(**model_params)

        # TTS: Text-to-Speech
        if self.category == "tts":
            self.client = TTSClient(type=self.type,config_path=self.config_path)
            return self.client(**model_params)

        # TTI: Text-to-Image
        if self.category == "tti":
            self.client = TTIClient(type=self.type,config_path=self.config_path)
            return self.client(**model_params)
        
        # ITT: There are 4 ways to provide an image       
        if self.category == "itt":
            self.client = ITTClient(type=self.type,config_path=self.config_path)
            messages = None

            if "messages" in model_params:
                messages = model_params.pop("messages", None)
                return self.client(messages=messages, **model_params)

            text = "Describe the image"
            if "text" in model_params:
                text = model_params.pop("text", None)

            def process_image(image, text):
                image_format = image.format
                buffered = BytesIO()
                image.save(buffered, format=image_format)
                image_base64 = base64.b64encode(buffered.getvalue()).decode()
                return [
                    {"role": "user", "content": [
                        {"type": "text", "text": text},
                        {"type": "image_url", "image_url": {
                            "url": f"data:image/{image_format};base64,"+image_base64
                        }
                        }
                    ]}
                ]

            if "file_path" in model_params:
                file_path = model_params.pop("file_path", None)
                with open(file_path, "rb") as f:
                    image = Image.open(f)
                    messages = process_image(image, text)
                    return self.client(messages=messages, **model_params)

            if "file" in model_params:
                file = model_params.pop("file", None)
                image = Image.open(file)
                messages = process_image(image, text)
                return self.client(messages=messages, **model_params)

            if "image" in model_params:
                image = model_params.pop("image", None)
                messages = process_image(image, text)
                return self.client(messages=messages, **model_params)

            if "image_url" in model_params:
                image_url = model_params.pop("image_url", None)
                messages = [
                    {"role": "user", "content": [
                        {"type": "text", "text": text},
                        {"type": "image_url", "image_url": image_url}
                    ]}
                ]
                return self.client(messages=messages, **model_params)


        if self.category == "index":
            raise Exception("The 'index' command has deprecated in GGG. Use RAGClientAsync instead.")
        
        if self.category == "retrieve":
            raise Exception("The 'retrieve' command has deprecated in GGG. Use RAGClientAsync instead.")
        
        raise Exception(f"Unknown category: {self.category}")



import logging
import os
from colorlog import ColoredFormatter

def getLogger(name):

    # Create a logger object
    logger = logging.getLogger(name)

    # Remove all handlers associated with the logger
    while logger.handlers:
        logger.handlers.pop()

    # Set a new log level
    logger.setLevel(os.environ.get("LOG_LEVEL", "WARNING"))

    # Create a console handler
    console_handler = logging.StreamHandler()

    # Set the format for the console handler using ColoredFormatter
    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)-8s%(reset)s %(message_log_color)s%(message)s%(reset)s",
        datefmt=None,
        reset=True,
        log_colors={
            'DEBUG':    'bg_purple,fg_black',
            'INFO':     'bg_green,fg_black',
            'WARNING':  'bg_yellow,fg_black',
            'ERROR':    'bg_red,fg_black',
            'CRITICAL': 'bg_red,fg_white',
        },
        secondary_log_colors={
            'message': {
                'DEBUG':    'fg_purple',
                'INFO':     'fg_green',
                'WARNING':  'fg_yellow',
                'ERROR':    'fg_red',
                'CRITICAL': 'fg_red',
            }
        },
        style='%'
    )

    # Add formatter to the console handler
    console_handler.setFormatter(formatter)

    # Add console handler to the logger
    logger.addHandler(console_handler)

    return logger


from gai.tools.Scraper import Scraper
from gai.common.logging import getLogger
logger = getLogger(__name__)

class Googler:

    def __init__(self):
        self.results = []
        self.links=[]
        self.text= ""

    def google(self, search_query, n_results=10, period=None):
        try:
            scraper=Scraper()
            period="d"
            #GOOGLE_SEARCH = f"https://www.google.com/search?q={search_query}&tbs=qdr:{period}"
            GOOGLE_SEARCH = f"https://www.google.com/search?q={search_query}"
            self.text,self.links = scraper.scrape(GOOGLE_SEARCH)
        except Exception as e:
            logger.error('Googler.google: Failed to extract links. error=',e)

        try:
            #Scrape top n search results
            dict_items = list(self.links.items())
            for key, value in dict_items:
                if (value.startswith("https://www.google.com/url?q=")):
                    value = value.split("https://www.google.com/url?q=")[1].split("&sa=U&")[0]
                    if (value.startswith("https://accounts.google.com")):
                        continue
                    if (value.startswith("https://support.google.com")):
                        continue
                    if (value.startswith("https://maps.google.com")):
                        continue
                    if (value.startswith("/search")):
                        continue
                    self.results.append({"title":key,"url":value})
                    
            return self.results
        except Exception as e:
            logger.error('Googler.google: error=',e)




from gai.lib.ttt.ChunkWrapper import ChunkWrapper
from gai.lib.ttt.OpenAIChunkWrapper import OpenAIChunkWrapper
from gai.lib.ttt.AnthropicChunkWrapper import AnthropicChunkWrapper
from gai.common.utils import get_lib_config
from gai.common.http_utils import http_post_async
from gai.common.generators_utils import chat_string_to_list, chat_list_to_string
from gai.common.errors import ApiException
from gai.common.logging import getLogger
logger = getLogger(__name__)
import json,os
from gai.lib.ClientBase import ClientBase
from dotenv import load_dotenv
load_dotenv()

class TTTClientAsync(ClientBase):

    def __init__(self, type, config_path=None):
        super().__init__(category_name="ttt", type=type, config_path=config_path)

    async def __call__(self, messages:str|list, stream:bool=True, **generator_params):

        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        # If generator is openai-gpt4 or claude2-100k, 
        # use the respective API client instead of generator service.
        if self.type == "openai":
            return await self.gpt_4(messages, stream=stream, **generator_params)
        if self.type == "gai":
            return await self.api(messages, stream=stream, **generator_params)

        raise Exception("Generator type not supported.")


    async def api(self, messages:list, stream:bool, **generator_params):
        #logger.debug(f'TTTClient.api: messages={messages}')

        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        if not messages:
            raise Exception("Messages not provided")


        data = { 
            "messages": messages,
            "stream": stream,
            **generator_params
        }

        def streamer(response):
            for chunk in response.iter_lines():
                yield ChunkWrapper(chunk)

        try:
            url = self._get_gai_url()
            response = await http_post_async(url, data)
        except ApiException as he:

            # Switch to long context
            if he.code == "context_length_exceeded":
                try:
                    url = self._get_gai_url("url")
                    response = await http_post_async(url, data)
                except Exception as e:
                    logger.error(f"TTTClient.api: gaigen error={e}")
                    raise e
            else:
                raise he
        except Exception as e:
            logger.error(f"TTTClient.api: gaigen error={e}")
            raise e

        if not stream:
            if response.json()["choices"][0]["message"]["tool_calls"]:
                response.decode = lambda: {
                    "type":"function",
                    "name": response.json()["choices"][0]["message"]["tool_calls"][0]["function"]["name"],
                    "arguments": response.json()["choices"][0]["message"]["tool_calls"][0]["function"]["arguments"],
                }
            elif response.json()["choices"][0]["message"]["content"]:
                response.decode = lambda: {
                    "type": "content",
                    "content": response.json()["choices"][0]["message"]["content"]
                }
            return response
        return streamer(response)

    # Call GPT-4 API
    async def gpt_4(self, messages:list, stream:bool, **generator_params):
        import os

        from openai import OpenAI

        # Try to get API KEY from gai.json
        OPENAI_API_KEY = generator_params.pop("OPENAI_API_KEY", None)
        if not OPENAI_API_KEY:
            # Then try to find it in environment variables
            if not os.environ.get("OPENAI_API_KEY"):
                raise Exception(
                    "OPENAI_API_KEY not found in environment variables")
        client = OpenAI(api_key=OPENAI_API_KEY)

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield OpenAIChunkWrapper(chunk)

        model = "gpt-4"
            
        response = await client.chat.completions.create(
            model=model,
            messages=messages,
            stream=stream,
            **generator_params
        )

        if not stream:
            if response.choices[0].message.tool_calls:
                response.decode = lambda: {
                    "type":"function",
                    "name": response.choices[0].message.tool_calls[0].function.name,
                    "arguments": response.choices[0].message.tool_calls[0].function.arguments,
                }
            elif response.choices[0].message.content:
                response.decode = lambda: {
                    "type": "content",
                    "content": response.choices[0].message.content
                }
            return response


        if not stream:
            response.decode = lambda: response.choices[
                0].message.content if response.choices[0].message.content else ""
            return response
        return streamer(response)




from gai.common.http_utils import http_post
from gai.common.utils import get_lib_config
from gai.lib.ClientBase import ClientBase
from gai.common.logging import getLogger
logger = getLogger(__name__)


class TTSClient(ClientBase):

    def __init__(self, type, config_path=None):
        super().__init__(category_name="tts", type=type, config_path=config_path)

    def __call__(self, input, stream=True, voice=None, language=None):
        if not input:
            raise Exception("The parameter 'input' is required.")

        if self.type == "openai":
            return self.openai_tts(input=input, voice=voice)
        if self.type == "gai":
            return self.gai_tts(input=input,voice=voice,stream=stream,language=language)

        raise Exception("Generator type not supported.")

    def gai_tts(self, input, voice, stream,language):
        data = {
            "input": input,
            "stream": stream,
            "voice": voice,
            "language": language
        }
        response = http_post(self._get_gai_url(), data)
        return response

    def openai_tts(self, input, voice):
        import os
        import openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()

        if not input:
            raise Exception("Missing input parameter")

        if not voice:
            voice = "alloy"

        response = client.audio.speech.create(
            model='tts-1', input=input, voice=voice)
        return response.content



import os
from gai.common.utils import get_lib_config

class ClientBase:

    def __init__(self, category_name, type, config_path=None):
        self.category_name = category_name
        self.config = get_lib_config(config_path)
        self.type = type

    def _get_gai_url(self):
        key = f"{self.category_name}-gai"
        config = self.config["generators"].get(key, None)
        if not config:
            raise Exception(f"Gai config does not exist. {key}")
        url = config.get("url",None)
        return url



from IPython.display import display, HTML
def print_colored(text, color):
    display(HTML(f'<font color="{color}">{text}</font>'))
def highlight(text):
    display(HTML(f'<font color="yellow">{text}</font>'))
def error(text):
    display(HTML(f'<font color="red">{text}</font>'))